# (c) 2012-2018 Dativa, all rights reserved
# -----------------------------------------
#  This code is licensed under MIT license (see license.txt for details)

import math
import itertools
import datetime
import pandas as pd
from dativa.tools.pandas import string_to_datetime, datetime_to_string


class ColumnStats:
    """A class that describes the stats for a column"""
    minimum = None
    maximum = None
    references = None

    def __init__(self, analyser, row, clean_threshold, outlier_threshold):
        self.analyser = analyser
        self.name = str(row["column"])
        self.sample_records = row["sample_records"]
        self.sample_blank = row["sample_blank"]
        self.sample_number = row["sample_number"]
        self.sample_string = row["sample_string"]
        self.sample_dates = row["sample_dates"]
        self.total_records = row["total_records"]
        self.total_unique = row["total_unique"]
        self.string_min_length = row["string_min_length"]
        self.string_max_length = row["string_max_length"]
        self.string_3std_min_length = row["string_3std_min_length"]
        self.string_3std_max_length = row["string_3std_max_length"]
        self.number_minimum = row["number_minimum"]
        self.number_maximum = row["number_maximum"]
        self.number_3std_minimum = row["number_3std_minimum"]
        self.number_3std_maximum = row["number_3std_maximum"]
        self.date_format = row["date_format"]
        self.date_minimum = row["date_minimum"]
        self.date_maximum = row["date_maximum"]
        self.date_3std_minimum = row["date_3std_minimum"]
        self.date_3std_maximum = row["date_3std_maximum"]
        self.decimal_places = row["decimal_places"]
        self.mode = row["mode"]

        # derived numbers
        self.sample_non_blanks = self.sample_records - self.sample_blank
        self.sample_percent_blank = self.sample_blank / self.sample_records
        if self.sample_non_blanks > 0:
            self.percent_string = self.sample_string / self.sample_non_blanks
            self.percent_date = self.sample_dates / self.sample_non_blanks
            self.percent_number = self.sample_number / self.sample_non_blanks
        else:
            self.percent_string = 0
            self.percent_date = 0
            self.percent_number = 0

        self.clean_threshold = clean_threshold
        self.outlier_threshold = outlier_threshold

    def _round(self, i, round_up):
        if i < 0:
            return - self._round(-i, not round_up)
        elif i < 1:
            return 0

        power_of_ten = int(math.pow(10, round(math.log10(i))))

        if power_of_ten == 1:
            if round_up:
                return 10
            else:
                return 1

        integer = int(i / power_of_ten)
        remainder = (i / power_of_ten) - integer

        if round_up and remainder > 0:
            return (integer + 1) * power_of_ten
        else:
            return integer * power_of_ten

    def _round_down(self, i):
        return self._round(i, False)

    def _round_up(self, i):
        return self._round(i, True)

    @staticmethod
    def _round_dates(min_val, max_val):
        if (max_val - min_val) > datetime.timedelta(days=31):
            return min_val.replace(month=1, day=1, hour=0, minute=0, second=0, microsecond=0), max_val.replace(month=12,
                                                                                                               day=31,
                                                                                                               hour=23,
                                                                                                               minute=59,
                                                                                                               second=59,
                                                                                                               microsecond=999999)
        elif (max_val - min_val) > datetime.timedelta(days=1):
            return min_val.replace(day=1, hour=0, minute=0, second=0, microsecond=0), max_val.replace(
                year=max_val.year + int(max_val.month / 12), month=(max_val.month + 1) % 12, day=1, hour=0, minute=0,
                second=0, microsecond=0) - datetime.timedelta(microseconds=1)
        elif (max_val - min_val) > datetime.timedelta(hours=1):
            return min_val.replace(hour=0, minute=0, second=0, microsecond=0), max_val.replace(hour=23, minute=59,
                                                                                               second=59,
                                                                                               microsecond=999999)
        else:
            return min_val.replace(minute=0, second=0, microsecond=0), max_val.replace(minute=59, second=59,
                                                                                       microsecond=999999)

    def _guess_min(self, minimum, min_3std):
        if min_3std * self.outlier_threshold < minimum:
            return min_3std
        else:
            return minimum

    def _guess_max(self, maximum, max_3std):
        if max_3std * self.outlier_threshold > maximum:
            return max_3std
        else:
            return maximum

    @property
    def type(self):
        if self.percent_number > self.clean_threshold:
            return "number"
        elif self.percent_date > self.clean_threshold and self.date_format != '':
            return "date"
        else:
            return "string"

    @property
    def is_unique(self):
        return self.total_unique / self.total_records > self.clean_threshold

    def _get_min_max(self):
        if self.type == "number":
            min_val = self._guess_min(self.number_minimum, self.number_3std_minimum)
            max_val = self._guess_max(self.number_maximum, self.number_3std_maximum)

            return self._round_down(min_val), self._round_up(max_val)

        if self.type == "date":
            min_diff = self.date_3std_minimum - self.date_minimum
            max_diff = self.date_maximum - self.date_3std_maximum
            std3_diff = self.date_3std_maximum - self.date_3std_minimum

            if min_diff > std3_diff * self.outlier_threshold:
                min_val = self.date_3std_minimum
            else:
                min_val = self.date_minimum

            if max_diff > std3_diff * self.outlier_threshold:
                max_val = self.date_3std_maximum
            else:
                max_val = self.date_maximum

            return self._round_dates(min_val, max_val)
        else:
            return (self._round_down(self._guess_min(self.string_min_length, self.string_3std_min_length)),
                    self._round_up(self._guess_max(self.string_max_length, self.string_3std_max_length)))

    def _check_min_max(self):
        if self.minimum is None or self.maximum is None:
            self.minimum, self.maximum = self._get_min_max()

    @property
    def min(self):
        self._check_min_max()
        return self.minimum

    @property
    def max(self):
        self._check_min_max()
        return self.maximum

    @property
    def has_blanks(self):
        return self.sample_percent_blank > (1 - self.clean_threshold)

    @property
    def fallback_mode(self):
        if self.type == "date":
            return "quarantine"
        elif self.type == "number":
            return "quarantine"
        elif self.type == "string":
            if self.number_references == 0:
                return "default"
            else:
                if self.number_references == self.total_unique:
                    return "quarantine"
                else:
                    return "closest"

    def _get_references(self):
        references = []
        if self.type == "string":
            # calculate reference rules
            maximum_records = self.total_records / self.analyser.min_occurences

            references = self.analyser.get_reference_list(self, maximum_records)

        return references

    def _check_references(self):
        if self.references is None:
            self.references = self._get_references()

    @property
    def reference_list(self):
        self._check_references()
        return self.references

    @property
    def number_references(self):
        self._check_references()
        return len(self.references)

    @property
    def default_value(self):
        if self.has_blanks or self.mode is None:
            return ""
        else:
            return self.mode


class DataFrameAnalyzer:
    """A class that contains functions to to analyse a passed DataFrame"""

    date_formats = ["%A, %B %d, %Y",
                    "%A, %B %d, %Y %I:%M:%S %p",
                    "%a, %d %b %Y %H:%M:%S %Z",
                    "%B, %Y",
                    "%d-%m-%Y %H:%M:%S"
                    "%d/%b/%Y:%H:%M:%S %z",
                    "%d/%m/%Y %H:%M:%S",
                    "%d/%m/%Y %H:%M",
                    "%d/%m/%Y",
                    "%H:%M:%S",
                    "%H:%M",
                    "%I:%M:%S %p",
                    "%I:%M:%S%p",
                    "%I:%M %p",
                    "%I:%M%p",
                    "%m/%d/%Y %I:%M:%S %p",
                    "%m/%d/%Y",
                    "%s",
                    "%Y%m%d %H%M%S",
                    "%Y%m%d %H:%M:%S",
                    "%Y%m%dT%H:%M:%S",
                    "%Y%m%dT%I:%M:%S %p",
                    "%Y%m%dT%H%M%S",
                    "%Y%m%dt%I%M%S %p",
                    "%Y-%m-%d %H:%M:%SZ",
                    "%Y-%m-%d %H:%M:%S",
                    "%Y-%m-%dT%H:%M:%S.000Z",
                    "%Y-%m-%dT%H:%M:%S.%fZ",
                    "%Y-%m-%dT%H:%M:%S.000",
                    "%Y-%m-%dT%H:%M:%S.%f",
                    "%Y-%m-%dT%I:%M:%S %p",
                    "%Y-%m-%dT%H:%M:%S",
                    "%Y/%m/%d %H:%M:%S",
                    "%Y/%m/%d",
                    "%Y:%m:%d %H:%M:%S"
                    ]

    def __init__(self,
                 df,
                 large_sample_size=5000,
                 small_sample_size=100,
                 clean_threshold=0.95,
                 outlier_threshold=2,
                 min_occurences=5):
        """Initiate a DataFrameAnalyser class, takes
        - df: the DataFrame to be analysed
        - large_sample_size: the maximum sample size to be taken for type-specific analysis
        - small_sample_size: the smaller sample size for inferring date format, decimal places
        """
        for d in df.dtypes:
            if d != object:
                raise Exception("DataFrame can only contain string columns, found type {0}".format(d))

        self.df = df
        self.large_sample_size = large_sample_size
        if self.large_sample_size > self.df.shape[0]:
            self.large_sample_size = self.df.shape[0]
        self.clean_threshold = clean_threshold
        self.outlier_threshold = outlier_threshold
        self.sample = self.df.sample(self.large_sample_size)
        self.min_occurences = min_occurences
        if small_sample_size > self.large_sample_size:
            self.small_sample_size = self.large_sample_size
        else:
            self.small_sample_size = small_sample_size
        self.small_sample = self.sample.sample(self.small_sample_size)
        self.columns = []

        for i, r in self._get_df_stats().iterrows():
            self.columns.append(ColumnStats(self, r, clean_threshold, outlier_threshold))

    @staticmethod
    def _is_number(s):
        try:
            float(s)
        except ValueError:
            return False
        else:
            return True

    def _get_number_string_blank(self, s):
        """Return the type of the passed string as either:
            b - blank
            n - number
            d - date
            s - string
        """
        if s == '' or s == 'nan':
            return 'b'
        if self._is_number(s):
            return "n"
        return "s"

    @staticmethod
    def _count(column, df, value):
        """return the number of records in the DataFrame's column with a specific value"""
        return df.loc[df[column] == value, column].count()

    @staticmethod
    def _std3_min_max(series):
        """returns the minimum and maximum values of the series within 3 standard deviations of the mean"""
        if series.shape[0] > 0:
            if series.shape[0] > 1:
                series = series[~((series - series.mean()).abs() > 3 * series.std())]
            return series.min(), series.max()
        return pd.Series(0), pd.Series(0)

    ##################
    # STRING STATISTICS
    ##################

    @staticmethod
    def _strlen(df, types, column):
        """returns the length of each row"""
        lengths = df.loc[types[column] != 'b', column].str.len()
        if lengths.shape[0] == 0:
            return pd.Series(0)
        else:
            return lengths

    def _get_string_stats(self, stats_df, types, df):
        """calculates summary statistics on all string rows"""
        minimum = stats_df["column"].map(lambda c: self._strlen(df, types, c).min())
        maximum = stats_df["column"].map(lambda c: self._strlen(df, types, c).max())
        min_3std, max_3std = zip(*stats_df["column"].map(lambda c: self._std3_min_max(self._strlen(df, types, c))))

        return minimum, maximum, min_3std, max_3std

    ##################
    # NUMBER STATISTICS
    ##################

    @staticmethod
    def _get_decimal_places(c):
        """returns the the number of decimal places in the string"""
        try:
            s = str(c)
            return len(s) - s.index('.') - 1
        except ValueError:
            return 0

    def _get_dp_frequency(self, column):
        """function to calculate the required decimal places to get the threshold fraction
        of the records included"""
        numbers = self.small_sample.loc[self.small_sample[column].map(lambda s: self._is_number(str(s))), column]
        if len(numbers) > 0:
            dp = numbers.map(lambda n: self._get_decimal_places(n))
            return int(dp.max())
        return 0

    @staticmethod
    def _number(df, types, column):
        """returns the numeric value of each row"""
        return pd.to_numeric(df.loc[types[column] == 'n', column])

    def _get_number_stats(self, stats_df, types, df):
        # basic stats
        minimum = stats_df["column"].map(lambda c: self._number(df, types, c).min())
        maximum = stats_df["column"].map(lambda c: self._number(df, types, c).max())
        min_3std, max_3std = zip(*stats_df["column"].map(lambda c: self._std3_min_max(self._number(df, types, c))))
        return (minimum,
                maximum,
                min_3std,
                max_3std)

    ##################
    # DATE STATISTICS
    ##################

    def _get_date_format(self, df, types, column):
        """get the date format string"""
        date_format = ""
        if len(df.loc[types[column] == 'd', column]) > 0:
            for f in self.date_formats:
                try:
                    timestamp = string_to_datetime(
                        df.loc[df.loc[types[column] == 'd', column].notnull().index, column].iloc[0], format=f)
                    string = df.loc[df.loc[types[column] == 'd', column].notnull().index, column].iloc[0]
                    if string == datetime_to_string(timestamp, f):
                        date_format = f
                        break
                except ValueError:
                    pass
        return date_format

    @staticmethod
    def _date(df, types, column, f):
        """return a date"""
        return string_to_datetime(df.loc[types[column] == 'd', column], f, errors="coerce")

    @staticmethod
    def _std3_date_min_max(series):
        """returns the minimum and maximum values of a series of dates within 3 standard deviations of the mean"""
        if series.shape[0] > 0:
            ts = series[series.notnull()].astype('int64')
            if ts.shape[0] > 1:
                ts = ts[~((ts - ts.mean()).abs() > 3 * ts.std())]
            return pd.to_datetime(ts.min(), unit='ns'), pd.to_datetime(ts.max(), unit='ns')
        return pd.np.nan, pd.np.nan

    def _get_date_stats(self, stats_df, types, df):
        """calculate the statistics for date columns"""
        minimum = stats_df[["column", "date_format"]].apply(lambda row: self._date(df, types, row[0], row[1]).min(),
                                                            axis=1)
        maximum = stats_df[["column", "date_format"]].apply(lambda row: self._date(df, types, row[0], row[1]).max(),
                                                            axis=1)
        min_3std, max_3std = zip(*stats_df[["column", "date_format"]].apply(
            lambda row: self._std3_date_min_max(self._date(df, types, row[0], row[1])), axis=1))

        return (minimum,
                maximum,
                min_3std,
                max_3std)

    def _get_mode(self, c):
        m = self.df.loc[self.df[c] != "", c].mode()
        if m.size == 0:
            return None
        else:
            return m[0]

    def _get_df_stats(self):
        """Returns a DataFrame that contains summary statistics for each
        column in the passed DataFrame. Some statistics are time intensive to calculate,
        so a sample is used based on the parameter large_sample_size"""

        # setup the DataFrame...
        stats_df = pd.DataFrame()

        # calculate the type of each field in the sample...
        types = self.sample.applymap(lambda x: self._get_number_string_blank(str(x)))

        # check the date formats
        types[self.sample[types == 'n'].applymap(lambda c: float(c) > 631152000 and float(c) < 4102444800)] = 'd'
        for c in types.columns:
            strings = self.small_sample.loc[types[c] == 's', c]
            if strings.shape[0] > 0:
                for f in self.date_formats:
                    try:
                        timestamp = string_to_datetime(strings, format=f, errors="raise").iloc[0]
                        if datetime_to_string(timestamp, f) == strings.iloc[0]:
                            types.loc[(types == 's').index, c] = 'd'
                            break
                    except ValueError:
                        pass

        # add the basic summary statistics...
        stats_df["column"] = types.columns
        stats_df["sample_records"] = types.shape[0]
        stats_df["sample_blank"] = stats_df["column"].map(lambda c: self._count(c, types, "b"))
        stats_df["sample_number"] = stats_df["column"].map(lambda c: self._count(c, types, "n"))
        stats_df["sample_string"] = stats_df["column"].map(lambda c: self._count(c, types, "s"))
        stats_df["sample_dates"] = stats_df["column"].map(lambda c: self._count(c, types, "d"))
        stats_df["total_records"] = self.df.shape[0]

        # add the unique statistics...
        stats_df["total_unique"] = stats_df["column"].map(lambda c: self.df[c].unique().shape[0])

        # get the most common value...
        stats_df["mode"] = stats_df["column"].map(lambda c: self._get_mode(c))

        # get the string stats...
        (stats_df["string_min_length"],
         stats_df["string_max_length"],
         stats_df["string_3std_min_length"],
         stats_df["string_3std_max_length"]) = self._get_string_stats(stats_df, types, self.sample)

        # get the date stats...
        stats_df["date_format"] = stats_df["column"].map(lambda c: self._get_date_format(self.small_sample, types, c))
        (stats_df["date_minimum"],
         stats_df["date_maximum"],
         stats_df["date_3std_minimum"],
         stats_df["date_3std_maximum"]) = self._get_date_stats(stats_df, types, self.sample)

        # get the number stats...
        (stats_df["number_minimum"],
         stats_df["number_maximum"],
         stats_df["number_3std_minimum"],
         stats_df["number_3std_maximum"]) = self._get_number_stats(stats_df, types, self.sample)

        stats_df["decimal_places"] = stats_df["column"].map(lambda c: self._get_dp_frequency(c))

        return stats_df

    def get_reference_list(self,
                           column,
                           max_records):
        """function to calculate the a reference list"""
        strings = self.df.loc[self.df[column.name].notnull().index, column.name]
        frequency = strings.value_counts()
        normalised_frequency = frequency[frequency >= self.min_occurences]
        number_normalised = normalised_frequency.shape[0]
        if (normalised_frequency.sum() > self.clean_threshold * frequency.sum() and
                number_normalised <= max_records):
            return normalised_frequency.index
        else:
            return []

    @staticmethod
    def _get_column_combinations(columns, max_cols):
        if len(columns) - 1 < max_cols:
            max_cols = len(columns) - 1
        for l in range(1, max_cols):
            for combination in itertools.combinations(columns, l):
                yield combination

    def get_unique_columns(self, threshold, max_cols):
        """This function searches for multiple column unique keys"""
        threshold_rows = self.sample.shape[0] * (1 - threshold)
        columns = self.sample.columns.values.tolist()
        for cols in self._get_column_combinations(columns, max_cols):
            number_duplicated = self.sample[self.sample.duplicated(subset=cols)].shape[0]
            if number_duplicated < threshold_rows:
                return list(cols)
        return []



# (c) 2012-2018 Dativa, all rights reserved
# -----------------------------------------
#  This code is licensed under MIT license (see license.txt for details)

import logging
from collections import OrderedDict
from dativa.tools.pandas import get_column_name
from .base import DefaultProfile, ScrubberTooManyRecordsError, ScrubberValidationError
from .validators import NumberValidator, StringValidator, DateValidator, SessionValidator, UniqueFieldsValidator, \
    LookupValidator
from .report import DefaultReportWriter

logger = logging.getLogger("dativa.scrubber")


class Scrubber:
    """ Utility class to process jobs

    Parameters
    ----------
    report_writer : ReportWriter class that is used to write out logs from
        any jobs run using the class
    profile : User profile sets limits to the number of records that can be
        processed in any job

    Example
    -------
    fp = FileProcessor()
    df = pd.read_csv(dirty_file)

    report = fp.run(df,
                    config={"rules": [
                     {
                         "rule_type": "String",
                         "field": "Name"
                         "params": {
                             "fallback_mode": "remove_record",
                             "maximum_length": 65536,
                             "minimum_length": 1
                         },
                     }]})

    for entry in report:
        print (entry)
        print (entry.df.describe())

    See also
    --------
    DataFrame.from_records : constructor from tuples, also record arrays
    DataFrame.from_dict : from dicts of Series, arrays, or dicts
    DataFrame.from_items : from sequence of (key, value) pairs
    pandas.read_csv, pandas.read_table, pandas.read_clipboard
    """

    validators = {
        "Lookup": LookupValidator,
        "Uniqueness": UniqueFieldsValidator,
        "Session": SessionValidator,
        "Date": DateValidator,
        "Number": NumberValidator,
        "String": StringValidator
    }

    def __init__(self, report_writer=None, profile=None):

        if report_writer is None:
            self._report_writer = DefaultReportWriter()
        else:
            self._report_writer = report_writer

        self.df = None

        if profile is None:
            self._profile = DefaultProfile()
        else:
            self._profile = profile

    def __del__(self):
        self.df = None

    def _extract_objects(self, obj, object_list=[]):
        """"
        Generator function to yield all of the objects in a nested dict or list
        that match the passed object list
        """
        if type(obj) != str:
            for key in obj:
                if type(obj[key]) == dict or type(obj[key]) == OrderedDict:
                    yield from self._extract_objects(obj[key], object_list)
                elif type(obj[key]) == list or type(obj[key]) == tuple:
                    for index in range(0, len(obj[key])):
                        yield from self._extract_objects(obj[key][index], object_list)
                else:
                    if key in object_list:
                        yield obj[key]

    def get_files_from_config(self, config, object_list=None):
        """
        Returns all of the files specified within a config files
        """
        if object_list is None:
            object_list = ["original_reference", "token_store"]
        return self._extract_objects(config, object_list)

    def _get_validator(self, rule, df_dict):
        """
        Returns an instance of the appropriate validator class depending on the
        rule type
        """
        try:
            return self.validators[rule["rule_type"]](rule["params"], self._report_writer, df_dict,
                                                      profile=self._profile)
        except KeyError:
            raise ScrubberValidationError("{name} is not a valid type of rule".format(name=rule["rule_type"]))

    def _run_single_column_rule(self, rule, column, df_dict):
        """
        Run the cleansing of the rule against a job and DataFrame

        Parameters
        ----------
        rule: an object containing the configuration of the rule to be run
        column: the name of the column on which to run the rule
        df_dict: a dictionary of DataFrame used when the rule configuration
            references an external file
        """

        # run the rule. this function will return the DF with clean values
        # appended as an additional column and a reference list for auto matching
        logger.info("Running rule {0} on column {1}".format(rule["rule_type"], column))
        validator = self._get_validator(rule, df_dict)
        validator.run_all(self.df, column)
        logger.info("Completed rule {0} on column {1}".format(rule["rule_type"], column))

        # now copy the new data into the original column
        if "append_results" not in rule or rule["append_results"] is False:
            try:
                # in the case of an optional column, `column` is a label, not an index
                self.df.iloc[:, column] = self.df.iloc[:, -1]
                self.df.drop(self.df.columns[-1], axis=1, inplace=True)
            except ValueError:
                # if that is the case then either the dataframe is unchanged, or we have quarantined it.
                pass

    def _run_multi_column_rule(self, rule, df_dict):
        """
        Run the cleansing of a multi-column rule against a job and DataFrame

        Parameters
        ----------
        rule: an object containing the configuration of the rule to be run
        df_dict: a dictionary of DataFrames used when the rule configuration
            references an external file
        """
        logger.info("Running multi-column rule {0}".format(rule["rule_type"]))
        validator = self._get_validator(rule, df_dict)
        validator.run_all(self.df)
        logger.info("Completed multi-column rule {0}".format(rule["rule_type"]))

    def run(self, df, config, df_dict={}):
        """
        Cleanses a DataFrame in-situ according to the passed configuration

        Parameters
        ----------
        df: the DataFrame to be cleansed
        config: a config dict containing the configuration
        df_dict: a dictionary of any DataFrame references in the configuration
        """
        self._report_writer.reset_report()
        self.df = df

        if self.df.shape[0] > self._profile.maximum_records:
            logger.info("Not processing job because too many records")
            raise ScrubberTooManyRecordsError(
                'There are too many records in the submitted file: please include no more '
                'than {0} records in your file'.format(self._profile.maximum_records))

        self.validate_config(config)

        # count the number of records and update the database

        original_total_columns = self.df.shape[1]
        for rule in config["rules"]:
            if self.df.shape[0] > 0:  # stop processing if all records have been deleted
                if "field" in rule and rule["field"] is not None:
                    try:
                        # if the 'field' is not in columns we cannot get_loc, but still need to pass on the name
                        # as we might be dealing with an optional column
                        column = df.columns.get_loc(get_column_name(self.df, rule["field"], original_total_columns))
                    except KeyError:
                        column = rule["field"]
                    self._run_single_column_rule(rule=rule,
                                                 column=column,
                                                 df_dict=df_dict)
                else:
                    self._run_multi_column_rule(rule=rule, df_dict=df_dict)

        return self._report_writer.get_report()

    def validate_config(self, config):
        """
        Checks that a passed configuration is valid and raises an exception if it is not

        Parameters
        ----------
        config: a config dict containing the configuration
        """

        try:
            for rule in config["rules"]:
                try:
                    self._get_validator(rule, {})
                except ScrubberValidationError as error:
                    if "field" in rule and rule["field"] is not None:
                        raise ScrubberValidationError("Error validating field {0}: {1}".format(rule["field"], error))
                    else:
                        raise ScrubberValidationError(
                            "Error validating multi-column rule: {1}".format(rule["field"], error))

        except KeyError:
            raise ScrubberValidationError("rules not present in config")

        return True



# (c) 2012-2018 Dativa, all rights reserved
# -----------------------------------------
#  This code is licensed under MIT license (see license.txt for details)

import datetime
import pandas as pd


class ReportEntry:
    """
    An individual report entry.

    Parameters
    ----------
    rule: the name of the rule that produced this entry
    field: the field being processed when the entry was produced
    df: the DataFrame containing the invalid records
    category: the type of entry, one of:
        HistoryCategory.IGNORED - the records were not altered
        HistoryCategory.REPLACED - the records were replaced with other values
        HistoryCategory.REMOVED - the records have been removed
        HistoryCategory.MODIFIED - the records have been modified
        HistoryCategory.INSERTED - new records have been inserted
    description: a textual description of the changes made (if any)
    """

    date = ""
    field = ""
    df = None
    category = ""
    description = ""
    number_records = ""

    def __init__(self, rule, field, df, category, description):
        self.date = datetime.datetime.now()
        self.field = field
        self.df = pd.DataFrame(df)
        self.number_records = df.shape[0]
        self.category = category
        self.description = description
        self.rule = rule

    def __str__(self):
        return "{0}, Field {1}({2}): #{3} {4}/|{5}".format(self.date,
                                                           self.field,
                                                           self.rule,
                                                           self.number_records,
                                                           self.category,
                                                           self.description)

    def get_log_dict(self):
        """
        Returns a dict containing the log entries
        """
        return {"date": self.date,
                "field": self.field,
                "rule": self.rule,
                "number_records": self.number_records,
                "category": self.category,
                "description": self.description}

    def get_records(self):
        """
        returns the DataFrame containing the values
        """
        return self.df


class DefaultReportWriter:
    """
    A default implementation of the ReportWriter that stores ReportEntries
    and returns them.
    """

    report = []

    def reset_report(self):
        self.report = []

    def log_history(self,
                    rule,
                    field,
                    df,
                    category,
                    description
                    ):
        """
        Logs a report entry
        """

        if isinstance(df, pd.Series):
            df = df.to_frame()

        self.report.append(ReportEntry(rule=rule,
                                       field=field,
                                       df=df,
                                       category=category,
                                       description=description
                                       ))

        return

    def get_report(self):
        """
        Returns an array of all ReportEntries created
        """
        return self.report



# (c) 2012-2018 Dativa, all rights reserved
# -----------------------------------------
#  This code is licensed under MIT license (see license.txt for details)

import logging
import pandas as pd
import numpy as np
from dativa.tools.pandas import get_column_name, get_unique_column_name
from .distance import get_closest_match, run_lookalike
from .base import BaseAddIn, HistoryType, FallbackType

logger = logging.getLogger("dativa.scrubber.fallback")


class BestMatchesAddIn(BaseAddIn):
    """
    AddIn to add functionality to a Validator that sets unknown values to the closest value
    """

    fields = {'attempt_closest_match': {
        "default": False, "validator": lambda x: x in [True, False]},
        'string_distance_threshold': {
            "default": 0.7, "validator": lambda x: (0 <= x <= 1)}}

    def _apply_best_matches(self, df, source_column, destination_column, reference_list, string_distance_threshold):
        """
        Copies the closest matching item from a reference list into the destination column
        in the current DataFrame. The reference list is matched against all values in the original
        DataFrame where the source column is not null and the destination column is.

        Parameters
        ----------
        df: the DataFrame
        source_column: the name of the source column in the DataFrame
        destination_column: the name of the destination column in the reference_list to use for best matches
        reference_list: a series contains the list of potential best matches
        string_distance_threshold: the threshold at which best matches
            are applied
        """

        if len(reference_list) > 0:

            source = get_column_name(df, source_column)
            destination = get_column_name(df, destination_column)

            logger.info("Applying best matches to rule {0} on column {1}".format(
                self.validator.rule_type, source_column))

            source_values = df.loc[df[destination].isnull(
            ) & df[source].notnull(), source].unique()

            if 0 < len(source_values) < self.validator.profile.maximum_records_closest_matches:
                source_values_df = pd.DataFrame(
                    source_values, columns=[source], dtype=pd.np.str)
                # create the possible substitutions objects in reverse order
                source_values_df["clean"], source_values_df["distance"] = get_closest_match(
                    source_values, reference_list.values)
                merged_df = df.loc[:, [source]].reset_index().merge(
                    source_values_df, how='left', left_on=source, right_on=source).set_index('index')
                merged_df.loc[merged_df["distance"] >
                              float(string_distance_threshold), "replace"] = merged_df["clean"]
                if merged_df[merged_df["replace"].notnull()].shape[0] > 0:
                    df.loc[merged_df["clean"].notnull(
                    ), destination] = merged_df["replace"]

                    self.report_writer.log_history(
                        self.validator.rule_type,
                        get_column_name(df, source_column),
                        df[merged_df["replace"].notnull(
                        )][[source, destination]],
                        HistoryType.REPLACED,
                        "Best match")

            else:
                logger.info("Skipped best matches for rule {0} on column {1}".format(
                    self.validator.rule_type, source_column))

    def run(self, df, column, original_column):

        if self.validator.reference_list is not None and self.validator.attempt_closest_match:
            # apply the best matches from the invalid records for now...
            self._apply_best_matches(df,
                                     original_column,
                                     column,
                                     self.validator.reference_list,
                                     self.validator.string_distance_threshold)


class LookalikeAddIn(BaseAddIn):
    """
    AddIn to add functionality to a Validator that sets unknown values to the closest vlaue according
    to a lookalike match
    """

    fields = {'lookalike_match': {
        "default": False, "validator": lambda x: x in [True, False]},
        'string_distance_threshold': {
            "default": 0.7, "validator": lambda x: (0 <= x <= 1)}}

    @staticmethod
    def _get_lookalike_matrix(df, cols_to_compare, target_field):
        """
        Return a numpy matrix suitable for lookalike matching containing:
        - the strings to be used for the lookalike match, which are concatenated
          strings in all the relevant columns to compare
        - the values of the target field
        - a blank column of int64
        - a blank column of strings
        """
        return np.vstack((df[cols_to_compare].T.apply(str).values,
                          df[target_field].apply(str).values,
                          np.zeros((1, df.shape[0]), dtype=np.int64),
                          np.zeros((1, df.shape[0]), dtype=np.str)
                          ))

    def _apply_lookalikes(self, df, source_column, destination_column, string_distance_threshold):
        """
        Copies the value from the closest matching row into the destination column.

        Parameters
        ----------
        df: the DataFrame
        source_column: the name of the source column in the DataFrame,
            this is ignored in the lookalike match
        destination_column: the name of the destination column
        string_distance_threshold: the threshold at which best matches
            are applied
        """

        logger.info("Applying lookalikes to rule {0} on column {1}".format(
            self.validator.rule_type, source_column))

        # prepare the columns....
        source = get_column_name(df, source_column)
        destination = get_column_name(df, destination_column)
        lookalike_column = get_unique_column_name(df, "lookalike")

        # check there are sufficient clean records for the match...
        number_to_match = df[df[destination].isnull()].shape[0]
        total_records = df.shape[0]

        number_possible_records = total_records - number_to_match
        if number_possible_records > self.validator.profile.lookalike_number_records:
            number_possible_records = self.validator.profile.lookalike_number_records

        if (number_to_match < self.validator.profile.maximum_records_lookalike and
                df.shape[1] > 1 and
                df.shape[0] < self.validator.profile.maximum_file_records_lookalike):

            # run the lookalike
            cols_to_compare = df.columns
            cols_to_compare = cols_to_compare.drop(destination)
            cols_to_compare = cols_to_compare.drop(source)
            df[lookalike_column] = run_lookalike(matrix=self._get_lookalike_matrix(df, cols_to_compare, destination),
                                                 rows_to_match=df[destination].isnull(
                                                 ).values,
                                                 top_n=number_possible_records,
                                                 threshold=float(string_distance_threshold))

            # update the clean column and log the history....
            df.loc[df[lookalike_column] != '',
                   destination] = df[lookalike_column]

            if df[df[lookalike_column] != ''].shape[0] > 0:
                self.report_writer.log_history(
                    self.validator.rule_type,
                    get_column_name(df, source_column),
                    df[df[lookalike_column] !=
                       ''][[source, destination]],
                    HistoryType.REPLACED,
                    "Lookalike match")

            # remove the lookalike column...
            df.drop(lookalike_column, axis=1, inplace=True)
        else:
            logger.info("Skipped lookalikes for rule {0} on column {1}".format(
                self.validator.rule_type, source_column))

    def run(self, df, column, original_column=None):

        if self.validator.lookalike_match:
            self._apply_lookalikes(df,
                                   original_column,
                                   column,
                                   self.validator.string_distance_threshold)


class FallbackAddIn(BaseAddIn):
    """
    AddIn to add functionality to a Validator to set unmatched values according to the fallback
    modes
    """

    fields = {'fallback_mode': {"default": None, "validator": lambda x: x in [
        FallbackType.USE_DEFAULT_VALUE, FallbackType.USE_INVALID_DATA, FallbackType.REMOVE_ENTRY]},
              'default_value': {
                  "default": "", "validator": lambda x: True}}

    def _apply_fallback_values(self, df, original_column, transformed_column, fallback_mode, default_value):
        """
        Sets any NULL values in the passed columns to the appropriate value

        Parameters
        ----------
        df: the DataFrame to run
        original_column, transformed_column: the column running on
        fallback_mode: one of three options:
            FallbackType.USE_DEFAULT_VALUE - replaces the Nulls with the specified default
            FallbackType.USE_INVALID_DATA - leaves the invalid data unchanged
            FallbackType.REMOVE_ENTRY - removes the invalid data and adds to quarantine
        default_value: the default value

        """
        rows = df.loc[df.iloc[:, -1].isnull()]

        logger.info("Applying fallback to {2}, mode {3} on rule {0} on column {1}".format(
            self.validator.rule_type, original_column, rows.shape[0], FallbackType.pretty(fallback_mode)))

        if rows.shape[0] > 0:

            if fallback_mode == FallbackType.USE_DEFAULT_VALUE:
                # set any empty rows in the final column to the default value
                df.loc[rows.index, transformed_column] = default_value
                self.report_writer.log_history(
                    self.validator.rule_type,
                    get_column_name(df, original_column),
                    df.loc[rows.index, [original_column, transformed_column]],
                    HistoryType.REPLACED,
                    FallbackType.pretty(fallback_mode)
                )

            elif fallback_mode == FallbackType.USE_INVALID_DATA:
                # set any empty rows in the final column to the  original value
                df.loc[rows.index, transformed_column] = df.loc[:, original_column]
                self.report_writer.log_history(
                    self.validator.rule_type,
                    get_column_name(df, original_column),
                    df.loc[rows.index, [original_column, transformed_column]],
                    HistoryType.IGNORED,
                    FallbackType.pretty(fallback_mode)
                )

            # remove nulls if replacement mode is to remove
            elif fallback_mode == FallbackType.REMOVE_ENTRY:
                self.report_writer.log_history(
                    self.validator.rule_type,
                    get_column_name(df, original_column),
                    df.loc[rows.index, original_column],
                    HistoryType.REMOVED,
                    FallbackType.pretty(fallback_mode)
                )

                df.drop(rows.index, inplace=True)

    def run(self, df, column, original_column=None):

        # put in any default values if present
        self._apply_fallback_values(df,
                                    original_column,
                                    column,
                                    self.validator.fallback_mode,
                                    self.validator.default_value)


class SkipBlankAddIn(BaseAddIn):
    """
    Adds functionality to a Validator to skip blanks
    """

    fields = {'skip_blank': {
        "default": False, "validator": lambda x: x in [True, False]}}

    @staticmethod
    def _copy_blanks(df, source_column, destination_column):
        """
        Copies all blanks from the source column to the destination column
        """
        df.loc[df.loc[:, source_column].isnull(), destination_column] = ''

        if df.dtypes[source_column] == object:
            df.loc[(df.loc[:, source_column] == ''), destination_column] = ''

    def run(self, df, column, original_column=None):

        if self.validator.skip_blank is True:
            self._copy_blanks(df, original_column, column)



#!/usr/bin/env python
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext

with open('README.md') as f:
    long_description = f.read()


def get_version():
    return open('version.txt', 'r').read().strip()


class NumpyBuildExt(build_ext):
    """build_ext command for use when numpy headers are needed."""

    def run(self):

        # Import numpy here, only when headers are needed
        import numpy  # noqa

        # Add numpy headers to include_dirs
        self.include_dirs.append(numpy.get_include())

        # Call original build_ext command
        build_ext.run(self)


setup(name='dativascrubber',
      version=get_version(),
      description='Dativa scrubber for automatic file cleansing',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='https://bitbucket.org/dativa4data/scrubber/',
      author='Dativa',
      author_email='hello@dativa.com',
      license='MIT',
      zip_safe=False,
      packages=['dativa.analyzer',
                'dativa.scrubber'],
      include_package_data=True,
      setup_requires=[
          'setuptools>=38.6.0',
          'wheel>=0.31.0',
          'numpy>=1.13.3'],
      install_requires=['setuptools>=18.0',
                        'pandas>=0.20.3',
                        'numpy>=1.13.3',
                        'python-levenshtein>=0.12.0',
                        'Cython>=0.27.3',
                        'pycrypto>=2.6.1',
                        'dativatools>=2.9.16'],
      scripts=['bin/fanalyzer'],
      test_suite='nose.collector',
      tests_require=['nose', 'coverage'],
      cmdclass={'build_ext': NumpyBuildExt},
      ext_modules=[Extension(name="dativa.scrubber.distance", sources=["dativa/scrubber/distance.pyx"])],
        classifiers=['Development Status :: 5 - Production/Stable',
                   'Intended Audience :: Developers',
                   'Topic :: Software Development :: Libraries',
                   'License :: OSI Approved :: MIT License',
                   'Programming Language :: Python :: 3.6'],
      keywords='dativa, data cleansing, data pipeline'
      )
